from .abc import (
    Model,
)
from .positions import (
    Position,
    GeometricPosition,
)
from .surfaces import (
    Surface,
    GeometricSurface,
    METRIC,
)
from .trips import (
    Trip,
    PlannedTrip,
)

from .jobs import (
    Job
)
from .vehicles import (
    Vehicle,
    Fleet,
)
from .routes import (
    Route,
)
from .plannings import (
    Planning,
)
from .results import (
    Result,
)



from time import time

from .abc import Dispatcher

from ..models import (
    Result,
)


class StaticDispatcher(Dispatcher):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def run(self) -> Result:
        loader = self.loader_cls()

        job = loader.job
        fleet = loader.fleet

        algorithm = self.algorithm_cls(fleet=fleet, job=job)

        start_time = time()
        planning = algorithm.optimize()
        end_time = time()
        computation_time = end_time - start_time

        result = Result(
            fleet=fleet,
            job=job,
            algorithm_cls=self.algorithm_cls,
            planning=planning,
            computation_time=computation_time,
        )

        storer = self.storer_cls(result=result)
        storer.store()
        return result



import unittest
from uuid import UUID

from jinete import (
    GeometricPosition,
)


class TestPositions(unittest.TestCase):

    def test_xy_position(self):
        position = GeometricPosition([3, 4])
        self.assertIsInstance(position.uuid, UUID)
        self.assertEqual(3, position[0])
        self.assertEqual(4, position[1])


if __name__ == '__main__':
    unittest.main()



import unittest

from uuid import UUID

from jinete import (
    GeometricSurface,
)
from .utils import generate_positions


class TestSurfaces(unittest.TestCase):

    def test_geometric_surface(self):
        positions = generate_positions(100)
        n = len(positions)
        surface = GeometricSurface(positions=positions)
        self.assertIsInstance(surface.uuid, UUID)
        self.assertEqual(n, len(surface.positions))

    def test_distance(self):
        positions = generate_positions(2)
        surface = GeometricSurface(positions=positions)

        a, b = tuple(positions)
        real_dist = ((a[0] - b[0]) ** 2 + (a[1] - b[1]) ** 2) ** 0.5

        dist = surface.distance(a, b)
        self.assertAlmostEqual(dist, real_dist)


if __name__ == '__main__':
    unittest.main()



from .models import (
    Position,
    GeometricPosition,
    Surface,
    GeometricSurface,
    METRIC,
    Trip,
    Job,
    PlannedTrip,
    Vehicle,
    Fleet,
    Route,
    Planning,
    Result,
)

from .loaders import (
    Loader,
    FileLoader,
    LoaderFormatter,
    HashCodeLoaderFormatter,
)

from .dispatchers import (
    Dispatcher,
    StaticDispatcher,
)

from .storers import (
    StorerFormatter,
    ColumnarStorerFormatter,
    HashCodeStorerFormatter,
    Storer,
    FileStorer,
    PromptStorer,
)

from .algorithms import (
    Algorithm,
    NaiveAlgorithm,
    GreedyAlgorithm,
)



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING
from time import time

from ..models import (
    Planning,
)

if TYPE_CHECKING:
    from ..models import (
        Fleet,
        Job,
        Route,
    )
logger = logging.getLogger(__name__)


class Algorithm(ABC):

    def __init__(self, fleet: Fleet, job: Job, *args, **kwargs):
        self.fleet = fleet
        self.job = job

    @abstractmethod
    def optimize(self) -> Planning:
        pass



from collections import namedtuple
from typing import (
    Set,
    Any,
    Dict,
    Callable, Tuple)

from .abc import (
    Model,
)
from .trips import (
    Trip,
)
from .routes import (
    Route,
)

OptimizationFunction = namedtuple('OptimizationFunction', {
    'direction': bool,
    'function': Callable[[Route], float],
    'description': str,
})

OPTIMIZATION_FUNCTIONS = {
    'DIAL_A_RIDE': OptimizationFunction(
        False,
        lambda route: sum(
            planned_trip.cost for planned_trip in route.planned_trips
        ),
        'Dial-a-Ride',
    ),
    'TAXI_SHARING': OptimizationFunction(
        False,
        lambda route: sum(
            planned_trip.duration for planned_trip in route.planned_trips if not planned_trip.capacity
        ),
        'Taxi-Sharing',
    )
}


class Job(Model):
    trips: Set[Trip]
    bonus: float
    optimization_function: OptimizationFunction

    def __init__(self, trips: Set[Trip], optimization_function: OptimizationFunction = None,
                 *args, **kwargs):
        if optimization_function is None:
            optimization_function = OPTIMIZATION_FUNCTIONS['DIAL_A_RIDE']

        self.trips = trips
        self.optimization_function = optimization_function

    def __iter__(self):
        yield from self.trips

    def as_dict(self) -> Dict[str, Any]:
        trips_str = ', '.join(str(trip) for trip in self.trips)
        dict_values = {
            'trips': f'{{{trips_str}}}',
            'optimization_function': self.optimization_function,
        }
        return dict_values



from __future__ import annotations

import logging
from itertools import product
from typing import Set, TYPE_CHECKING

from .abc import Algorithm
from ..models import (
    Route,
    Fleet,
    Job,
    Planning,
)

if TYPE_CHECKING:
    from ..models import (
        Trip,
    )

logger = logging.getLogger(__name__)


class Crosser(object):

    def __init__(self, fleet: Fleet, job: Job):
        self.routes = set(Route(vehicle) for vehicle in fleet.vehicles)
        self.trips = job.trips
        self.done_trips = set()

    def __iter__(self):
        for route, trip in (product(self.atractive_routes, self.pending_trips)):
            if self.completed:
                break
            logger.debug(f'Yielding ({route}, {trip})...')
            yield route, trip
        return None

    @property
    def atractive_routes(self) -> Set[Route]:
        routes = set(route for route in self.routes if len(route.planned_trips) > 0)
        empty_route = next((route for route in self.routes if len(route.planned_trips) == 0), None)
        if empty_route is not None:
            routes.add(empty_route)
        return routes

    def mark_as_done(self, trip: Trip):
        self.done_trips.add(trip)

    def mark_as_undone(self, trip: Trip):
        self.done_trips.remove(trip)

    @property
    def pending_trips(self):
        return self.trips - self.done_trips

    @property
    def completed(self) -> bool:
        return len(self.pending_trips) == 0

    def update_ranking(self):
        ranking = list()
        for route, trip in self:
            planned_trip = route.feasible_trip(trip)
            if planned_trip is None:
                continue
            ranking.append(planned_trip)
        ranking.sort(key=lambda pt: pt.scoring)

    def get_best(self):
        best_planned_trip = None
        for route, trip in self:
            planned_trip = route.feasible_trip(trip)
            if planned_trip is None:
                continue

            if best_planned_trip is None or planned_trip.scoring < best_planned_trip.scoring:
                best_planned_trip = planned_trip

        return best_planned_trip


class GreedyAlgorithm(Algorithm):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.crosser_cls = Crosser

    def build_crosser(self) -> Crosser:
        return self.crosser_cls(self.fleet, self.job)

    def optimize(self) -> Planning:
        logger.info('Optimizing...')
        crosser = self.build_crosser()

        while not crosser.completed:
            planned_trip = crosser.get_best()
            if not planned_trip:
                break
            route = planned_trip.route
            route.append_planned_trip(planned_trip)
            crosser.mark_as_done(planned_trip.trip)

        for route in crosser.routes:
            route.finish()
        planning = Planning(crosser.routes)
        logger.info('Optimized!')
        return planning



from __future__ import annotations

import logging
from dataclasses import dataclass, field
from sys import maxsize
from typing import (
    TYPE_CHECKING,
    Optional,
)

from uuid import (
    uuid4,
)

if TYPE_CHECKING:
    from uuid import (
        UUID,
    )
    from .positions import (
        Position,
    )
    from .routes import (
        Route,
    )

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class Trip(object):
    identifier: str
    origin: Position
    destination: Position

    earliest: float = field(default=0.0)
    timeout: Optional[float] = field(default=None)

    load_time: float = field(default=0.0)

    capacity: int = field(default=1)
    uuid: UUID = field(default_factory=uuid4)

    @staticmethod
    def build_empty(origin: Position, destination: Position) -> 'Trip':
        return Trip('EMPTY', origin, destination, capacity=0)

    @property
    def latest(self) -> float:
        if self.timeout is None:
            return maxsize
        return self.earliest + self.timeout

    @property
    def empty(self) -> bool:
        return self.capacity == 0

    @property
    def distance(self) -> float:
        return self.origin.distance_to(self.destination)

    def duration(self, now: float):
        return self.origin.time_to(self.destination, now)


@dataclass(frozen=True)
class PlannedTrip(object):
    route: Route
    trip: Trip
    collection_time: float
    delivery_time: float

    @staticmethod
    def build_empty(route, collection_time, delivery_time, *args, **kwargs) -> 'PlannedTrip':
        trip = Trip.build_empty(*args, **kwargs)
        return PlannedTrip(route, trip, collection_time, delivery_time)

    @property
    def origin(self) -> Position:
        return self.trip.origin

    @property
    def destination(self) -> Position:
        return self.trip.destination

    @property
    def duration(self) -> float:
        return self.delivery_time - self.collection_time

    @property
    def capacity(self):
        return self.trip.capacity

    @property
    def scoring(self):
        return self.collection_time - self.route.last_time

    @property
    def feasible(self) -> bool:
        return self.trip.earliest <= self.collection_time and self.delivery_time <= self.trip.latest

    @property
    def empty(self) -> bool:
        return self.trip.empty



import unittest
from uuid import UUID

import jinete as jit

from .utils import (
    generate_one_vehicle,
    generate_one_route,
)


class TestRoutes(unittest.TestCase):

    def test_route(self):
        planned_trips = list()
        vehicle = generate_one_vehicle()
        route = jit.Route(vehicle, planned_trips)
        self.assertIsInstance(route.uuid, UUID)
        self.assertEqual(planned_trips, route.planned_trips)
        self.assertEqual(vehicle, route.vehicle)

    def test_feasible_route(self):
        route = generate_one_route(True)
        self.assertTrue(route.feasible)

    def test_not_feasible_route(self):
        route = generate_one_route(False)
        self.assertFalse(route.feasible)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from .abc import StorerFormatter


class HashCodeStorerFormatter(StorerFormatter):
    def __init__(self, remove_empty_routs: bool = True, *args, **kwargs):
        super().__init__(
            remove_empty_routs=remove_empty_routs,
            *args, **kwargs,
        )

    @staticmethod
    def route_to_str(route) -> str:
        loaded_trips = route.loaded_trips
        trips_str = ' '.join(trip.identifier for trip in loaded_trips)
        return f'{len(loaded_trips)} {trips_str}'

    def format(self) -> str:
        result = str()
        result += '\n'.join(self.route_to_str(route) for route in self.routes)
        return result



from __future__ import annotations
import logging
from abc import (
    ABC,
    abstractmethod)

from math import sqrt
from typing import (
    TYPE_CHECKING,
)

from uuid import (
    uuid4,
)

from .abc import (
    Model,
)
from .positions import (
    GeometricPosition,
)

if TYPE_CHECKING:
    from typing import (
        Set,
        Any,
        Dict,
        Callable,
    )
    from uuid import (
        UUID,
    )
    from .positions import (
        Position,
    )

logger = logging.getLogger(__name__)


class Surface(Model, ABC):
    uuid: UUID

    def __init__(self, positions: Set[Position] = None, uuid: UUID = None, *args, **kwargs):
        if uuid is None:
            uuid = uuid4()
        if positions is None:
            positions = set()
        self.uuid = uuid
        self.positions = positions

    def get_or_create_position(self, *args, **kwargs) -> Position:
        position = self._build_position(*args, **kwargs)
        self.positions.add(position)
        return position

    @abstractmethod
    def _build_position(self, *args, **kwargs):
        pass

    @abstractmethod
    def distance(self, position_a: Position, position_b: Position) -> float:
        pass

    @abstractmethod
    def time(self, position_a: Position, position_b: Position, now: float) -> float:
        pass

    def as_dict(self) -> Dict[str, Any]:
        positions_str = ', '.join(str(position) for position in self.positions)
        dict_values = {
            'positions': f'{{{positions_str}}}'
        }
        return dict_values


METRIC = {
    'EUCLIDEAN': lambda a, b: sqrt(sum(pow(a_i - b_i, 2) for a_i, b_i in zip(a, b))),
    'MANHATTAN': lambda a, b: sum(abs(a_i - b_i) for a_i, b_i in zip(a, b)),
}


class GeometricSurface(Surface):
    positions: Set[GeometricPosition]

    def __init__(self, metric: Callable[[Any, Any], float] = None, *args, **kwargs):
        if metric is None:
            metric = METRIC['EUCLIDEAN']
        super().__init__(*args, **kwargs)
        self.metric = metric

    def _build_position(self, *args, **kwargs):
        return GeometricPosition(surface=self, *args, **kwargs)

    def distance(self, position_a: GeometricPosition, position_b: GeometricPosition, auto_add: bool = False) -> float:
        if position_a not in self.positions:
            logger.warning(f'"position_a"="{position_a}" is not on "self.positions".')
            if auto_add:
                self.positions.add(position_a)

        if position_b not in self.positions:
            logger.warning(f'"position_b"="{position_b}" is not on "self.positions".')
            if auto_add:
                self.positions.add(position_b)
        return self.metric(position_a, position_b)

    def time(self, position_a: GeometricPosition, position_b: GeometricPosition, now: float) -> float:
        return self.distance(position_a, position_b)



from __future__ import annotations

from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    TYPE_CHECKING,
)
from ..storers import (
    NaiveStorer,
)

if TYPE_CHECKING:
    from typing import (
        Type,
    )
    from ..loaders import (
        Loader,
    )
    from ..algorithms import (
        Algorithm,
    )
    from ..storers import (
        Storer,
    )


class Dispatcher(ABC):

    def __init__(self, loader_cls: Type[Loader], algorithm_cls: Type[Algorithm], storer_cls: Type[Storer] = None):
        if storer_cls is None:
            storer_cls = NaiveStorer
        self.loader_cls = loader_cls
        self.algorithm_cls = algorithm_cls
        self.storer_cls = storer_cls

    @abstractmethod
    def run(self):
        pass



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from .models import (
        Route,
        Trip,
    )


class JineteException(Exception):

    def __init__(self, message: str):
        super().__init__(message)


class PlannedTripNotFeasibleException(JineteException):

    def __init__(self, route: Route, trip: Trip):
        self.route = route
        self.trip = trip
        message = f'The {trip} is not feasible at the end of {route}.'
        super().__init__(message=message)



from abc import ABC, abstractmethod
from typing import Any, Dict


class Model(ABC):
    def __repr__(self):
        return self._print(self.as_dict())

    def _print(self, values):
        values = ', '.join(f'{key}={value}' for key, value in values.items())
        return f'{self.__class__.__name__}({values})'

    @abstractmethod
    def as_dict(self) -> Dict[str, Any]:
        pass



from __future__ import annotations

import logging
from typing import (
    TYPE_CHECKING,
    Any, Dict, Optional)
from uuid import (
    uuid4,
)
from .abc import (
    Model,
)
from .trips import (
    PlannedTrip,
)

if TYPE_CHECKING:
    from typing import (
        Tuple,
        Iterable,
        List,
    )
    from uuid import (
        UUID,
    )
    from .vehicles import (
        Vehicle,
    )
    from .trips import (
        Trip,
    )
    from .positions import (
        Position,
    )

logger = logging.getLogger(__name__)


class Route(Model):
    planned_trips: List[PlannedTrip]
    vehicle: Vehicle
    uuid: UUID

    def __init__(self, vehicle: Vehicle, planned_trips: Iterable[PlannedTrip] = None, uuid=None):
        if planned_trips is None:
            planned_trips = tuple()
        if uuid is None:
            uuid = uuid4()

        self.vehicle = vehicle
        self.planned_trips = list(planned_trips)
        self.uuid = uuid

    @property
    def feasible(self) -> bool:
        if len(self.planned_trips) > 0:
            if not self.first_trip.origin == self.vehicle.initial:
                return False
            if not self.vehicle.earliest <= self.first_planned_trip.collection_time:
                return False
            if not self.last_trip.destination == self.vehicle.final:
                return False
            if not self.last_planned_trip.delivery_time <= self.vehicle.latest:
                return False

        for planned_trip in self.planned_trips:
            if not planned_trip.feasible:
                return False
        return True

    @property
    def loaded(self):
        return len(self.planned_trips) > 0

    @property
    def trips(self) -> Tuple[Trip]:
        return tuple(planned_trip.trip for planned_trip in self.planned_trips)

    @property
    def loaded_planned_trips(self) -> Tuple[PlannedTrip]:
        return tuple(planned_trip for planned_trip in self.planned_trips if not planned_trip.empty)

    @property
    def loaded_trips(self) -> Tuple[Trip]:
        return tuple(planned_trip.trip for planned_trip in self.planned_trips if not planned_trip.empty)

    @property
    def first_planned_trip(self) -> PlannedTrip:
        return min(self.planned_trips, key=lambda pt: pt.collection_time)

    @property
    def first_trip(self) -> Trip:
        return self.first_planned_trip.trip

    @property
    def last_planned_trip(self) -> PlannedTrip:
        return max(self.planned_trips, key=lambda pt: pt.delivery_time)

    @property
    def last_trip(self) -> Trip:
        return self.last_planned_trip.trip

    @property
    def duration(self) -> float:
        return self.last_planned_trip.delivery_time - self.first_planned_trip.collection_time

    @property
    def last_position(self) -> Position:
        if len(self.planned_trips) == 0:
            return self.vehicle.initial
        return self.last_trip.destination

    @property
    def last_time(self) -> float:
        if len(self.planned_trips) == 0:
            return self.vehicle.earliest
        return self.last_planned_trip.delivery_time

    @property
    def vehicle_uuid(self) -> Optional[UUID]:
        if self.vehicle is None:
            return None
        return self.vehicle.uuid

    def as_dict(self) -> Dict[str, Any]:
        return {
            'uuid': self.uuid,
            'vehicle_uuid': self.vehicle_uuid,
        }

    def feasible_trip(self, trip: Trip) -> Optional[PlannedTrip]:
        time_to_go = self.last_position.distance_to(trip.origin)
        time_to_travel = trip.duration(time_to_go)
        trip_start_time = max(self.last_time + time_to_go, trip.earliest)
        trip_finish_time = trip_start_time + time_to_travel
        if not trip_finish_time <= trip.latest:
            return None
        time_to_return = trip.destination.time_to(self.vehicle.final, trip_finish_time)
        vehicle_finish_time = trip_finish_time + time_to_return
        if not vehicle_finish_time <= self.vehicle.latest:
            return None
        return PlannedTrip(self, trip, trip_start_time, trip_finish_time)

    def _append_empty_planned_trip(self, destination: Position) -> PlannedTrip:
        planned_trip = PlannedTrip.build_empty(
            route=self,
            collection_time=self.last_time,
            delivery_time=self.last_time + self.last_position.time_to(destination, self.last_time),
            origin=self.last_position,
            destination=destination,
        )
        self.append_planned_trip(planned_trip)
        return planned_trip

    def _append_finish_planned_trip(self) -> PlannedTrip:
        return self._append_empty_planned_trip(self.vehicle.final)

    def finish(self):
        if self.loaded:
            self._append_finish_planned_trip()

    def append_planned_trip(self, planned_trip: PlannedTrip):
        if not self.last_position.is_equal(planned_trip.origin):
            self._append_empty_planned_trip(planned_trip.origin)
        # if not planned_trip.feasible:
        #     raise PlannedTripNotFeasibleException(self, planned_trip.trip)
        self.planned_trips.append(planned_trip)
        logger.info(f'Append trip "{planned_trip.trip.identifier}" identifier to route "{self.uuid}".')



import unittest
from sys import maxsize
from uuid import UUID

import jinete as jit
from .utils import (
    generate_one_position,
)


class TestVehicles(unittest.TestCase):

    def test_vehicle(self):
        initial = generate_one_position()
        vehicle = jit.Vehicle(initial)

        self.assertEqual(1, vehicle.capacity)
        self.assertEqual(vehicle.initial, initial)
        self.assertEqual(vehicle.final, vehicle.initial)
        self.assertEqual(vehicle.earliest, 0)
        self.assertEqual(vehicle.latest, maxsize)
        self.assertIsNone(vehicle.timeout)
        self.assertIsInstance(vehicle.uuid, UUID)

    def test_vehicle_with_capacity(self):
        capacity = 3
        initial = generate_one_position()
        vehicle = jit.Vehicle(initial, capacity=capacity)

        self.assertEqual(vehicle.capacity, capacity)
        self.assertEqual(vehicle.initial, initial)
        self.assertEqual(vehicle.final, vehicle.initial)
        self.assertEqual(vehicle.earliest, 0)
        self.assertEqual(vehicle.latest, maxsize)
        self.assertIsNone(vehicle.timeout)
        self.assertIsInstance(vehicle.uuid, UUID)

    def test_vehicle_with_final(self):
        capacity = 3
        initial = generate_one_position()
        final = generate_one_position()
        vehicle = jit.Vehicle(initial, capacity=capacity, final=final)

        self.assertEqual(vehicle.capacity, capacity)
        self.assertEqual(vehicle.initial, initial)
        self.assertEqual(vehicle.final, final)
        self.assertEqual(vehicle.earliest, 0)
        self.assertEqual(vehicle.latest, maxsize)
        self.assertIsNone(vehicle.timeout)
        self.assertIsInstance(vehicle.uuid, UUID)

    def test_vehicle_with_earliest(self):
        initial = generate_one_position()
        earliest = 3600

        vehicle = jit.Vehicle(initial, earliest=earliest)

        self.assertEqual(vehicle.capacity, 1, )
        self.assertEqual(vehicle.initial, initial)
        self.assertEqual(vehicle.final, vehicle.initial)
        self.assertEqual(vehicle.earliest, earliest)
        self.assertEqual(vehicle.latest, maxsize)
        self.assertIsNone(vehicle.timeout)
        self.assertIsInstance(vehicle.uuid, UUID)

    def test_vehicle_with_timeout(self):
        initial = generate_one_position()
        earliest = 1800
        timeout = 3600

        vehicle = jit.Vehicle(initial, earliest=earliest, timeout=timeout)

        self.assertEqual(vehicle.capacity, 1)
        self.assertEqual(vehicle.initial, initial)
        self.assertEqual(vehicle.final, vehicle.initial)
        self.assertEqual(vehicle.earliest, earliest)
        self.assertEqual(vehicle.latest, earliest + timeout)
        self.assertEqual(vehicle.timeout, timeout)
        self.assertIsInstance(vehicle.uuid, UUID)


if __name__ == '__main__':
    unittest.main()



from pathlib import Path

from .abc import Storer


class FileStorer(Storer):

    def __init__(self, file_path: Path, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.file_path = file_path

    def store(self) -> None:
        with self.file_path.open('w') as file:
            text = self.formatted_planning()
            file.write(text)



from abc import ABC, abstractmethod
from typing import Any

from ...models import (
    Fleet,
    Job,
    Surface,
)


class LoaderFormatter(ABC):
    def __init__(self, data: Any):
        self.data = data

    @abstractmethod
    def fleet(self, *args, **kwargs) -> Fleet:
        pass

    @abstractmethod
    def job(self, *args, **kwargs) -> Job:
        pass

    @abstractmethod
    def surface(self, *args, **kwargs) -> Surface:
        pass



import unittest
from sys import maxsize
from uuid import UUID

import jinete as jit
from .utils import (
    generate_one_position,
)


class TestTrips(unittest.TestCase):

    def test_trip(self):
        identifier = str()
        origin = generate_one_position()
        destination = generate_one_position()
        earliest = 0

        trip = jit.Trip(identifier=identifier, origin=origin, destination=destination, earliest=earliest)

        self.assertEqual(origin, trip.origin)
        self.assertEqual(destination, trip.destination)
        self.assertEqual(earliest, trip.earliest)
        self.assertIsNone(trip.timeout)
        self.assertEqual(trip.latest, maxsize)
        self.assertEqual(0, trip.load_time)
        self.assertEqual(1, trip.capacity)
        self.assertIsInstance(trip.uuid, UUID)

    def test_trip_with_capacity(self):
        identifier = str()
        capacity = 3
        origin = generate_one_position()
        destination = generate_one_position()
        earliest = 0

        trip = jit.Trip(identifier=identifier, origin=origin, destination=destination, earliest=earliest,
                        capacity=capacity)

        self.assertEqual(origin, trip.origin)
        self.assertEqual(destination, trip.destination)
        self.assertEqual(earliest, trip.earliest)
        self.assertIsNone(trip.timeout)
        self.assertEqual(trip.latest, maxsize)
        self.assertEqual(0, trip.load_time)
        self.assertEqual(capacity, trip.capacity)
        self.assertIsInstance(trip.uuid, UUID)

    def test_trip_with_timeout(self):
        identifier = str()
        origin = generate_one_position()
        destination = generate_one_position()
        earliest = 0
        timeout = 3600
        trip = jit.Trip(identifier=identifier, origin=origin, destination=destination, earliest=earliest,
                        timeout=timeout)

        self.assertEqual(origin, trip.origin)
        self.assertEqual(destination, trip.destination)
        self.assertEqual(earliest, trip.earliest)
        self.assertEqual(timeout, trip.timeout)
        self.assertEqual(earliest + timeout, trip.latest)
        self.assertEqual(0, trip.load_time)
        self.assertEqual(1, trip.capacity)
        self.assertIsInstance(trip.uuid, UUID)

    def test_trip_with_load_time(self):
        identifier = str()
        origin = generate_one_position()
        destination = generate_one_position()
        earliest = 0
        load_time = 1800
        trip = jit.Trip(identifier=identifier, origin=origin, destination=destination, earliest=earliest,
                        load_time=load_time)

        self.assertEqual(origin, trip.origin)
        self.assertEqual(destination, trip.destination)
        self.assertEqual(earliest, trip.earliest)
        self.assertIsNone(trip.timeout)
        self.assertEqual(trip.latest, maxsize)
        self.assertEqual(load_time, trip.load_time)
        self.assertEqual(1, trip.capacity)
        self.assertIsInstance(trip.uuid, UUID)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import logging
from dataclasses import (
    dataclass,
    field,
)
from typing import (
    TYPE_CHECKING,
    Set,
    Optional)
from uuid import (
    uuid4,
)

if TYPE_CHECKING:
    from .routes import (
        Route,
    )
    from uuid import (
        UUID,
    )
    from .surfaces import (
        Surface,
    )

logger = logging.getLogger(__name__)


class Planning(object):
    routes: Set[Route]
    computation_time: float
    uuid: UUID

    def __init__(self, routes: Set[Route], uuid: UUID = None):
        if uuid is None:
            uuid = uuid4()

        self.routes = routes
        self.uuid = uuid

    @property
    def loaded_routes(self):
        return set(route for route in self.routes if route.loaded)



from __future__ import annotations
from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    TYPE_CHECKING,
)
from ..models import (
    Fleet,
    Job,
    Surface,
)
from .formatters import (
    HashCodeLoaderFormatter,
)

if TYPE_CHECKING:
    from typing import (
        Type,
    )
    from .formatters import (
        LoaderFormatter,
    )


class Loader(ABC):

    def __init__(self, formatter_cls: Type[LoaderFormatter] = None):
        if formatter_cls is None:
            formatter_cls = HashCodeLoaderFormatter
        self.formatter_cls = formatter_cls
        self.formatter: LoaderFormatter = None

    @property
    @abstractmethod
    def fleet(self) -> Fleet:
        pass

    @property
    @abstractmethod
    def job(self) -> Job:
        pass

    @property
    @abstractmethod
    def surface(self) -> Surface:
        pass



from itertools import chain
from typing import List

from .abc import StorerFormatter

from ...models import (
    Route,
    Vehicle,
    PlannedTrip,
)


class ColumnarStorerFormatter(StorerFormatter):

    def vehicle_to_str(self, vehicle: Vehicle) -> List[str]:
        return [
            f'ID: "{vehicle.uuid}"',
            f'Initial:  {vehicle.initial}',
            f'Final:    {vehicle.final}',
            f'Earliest: {vehicle.earliest:7.2f}',
            f'Latest:   {vehicle.latest:7.2f}',
            f'Capacity: {vehicle.capacity:7.2f}',
        ]

    def planned_trip_to_str(self, planned_trip: PlannedTrip) -> List[str]:
        return [
            '\t'.join((
                f'ID: {planned_trip.trip.identifier:6}',
                f'Position: {planned_trip.origin} to {planned_trip.destination}',
                f'Duration: {planned_trip.duration:7.02f}',
                f'Time: {planned_trip.collection_time:8.02f} to {planned_trip.delivery_time:8.02f}',
                f'Load: {planned_trip.capacity}',
            )),
        ]

    def route_to_str(self, route: Route) -> List[str]:
        rows = chain.from_iterable(self.planned_trip_to_str(planned_trip) for planned_trip in route.planned_trips)
        return [
            f'Vehicle: ',
            *(f'\t{row}' for row in self.vehicle_to_str(route.vehicle)),
            f'Planned Trips: "{len(route.planned_trips)}"',
            *(f'\t{row}' for row in rows)
        ]

    def format(self) -> str:
        rows = chain.from_iterable(self.route_to_str(route) for route in self.planning.routes)
        return '\n'.join((
            f'Planning UUID: "{self.planning.uuid}"',
            f'Routes count: "{len(self.planning.routes)}"',
            f'Routes: ',
            '\n'.join(f'\t{row}' for row in rows),
            f'Computation time: "{self.result.computation_time:0.4f}" seconds',
            f'Coverage Rate: "{self.result.coverage_rate}"',
        ))



import logging

from ...models import (
    Fleet,
    Job,
    Surface,
    GeometricSurface,
    METRIC,
    Vehicle,
    Trip,
)
from .abc import (
    LoaderFormatter,
)

logger = logging.getLogger(__name__)


class HashCodeLoaderFormatter(LoaderFormatter):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # TODO: Set bonus and number of steps (Config object)
        #

    def fleet(self, surface: Surface, *args, **kwargs) -> Fleet:
        row = self.data[0]
        n, timeout, capacity = row[2], row[5], 1

        initial = surface.get_or_create_position([0, 0])
        vehicles = set(Vehicle(initial, capacity=capacity, timeout=timeout) for _ in range(n))
        fleet = Fleet(vehicles)
        logger.info(f'Created fleet!')
        return fleet

    def job(self, surface: Surface, *args, **kwargs) -> Job:
        rows = self.data[1:]
        trips = set(self._build_trip(surface, str(i), *row) for i, row in enumerate(rows))
        job = Job(trips, *args, **kwargs)
        logger.info(f'Created job!')
        return job

    def _build_trip(self, surface: Surface, identifier: str, x1: float, y1: float, x2: float, y2: float,
                    earliest: float, latest: float) -> Trip:
        origin = surface.get_or_create_position([x1, y1])
        destination = surface.get_or_create_position([x2, y2])
        timeout = latest - earliest
        trip = Trip(identifier, origin, destination, earliest, timeout)
        logger.info(f'Created trip!')
        return trip

    def surface(self, *args, **kwargs) -> Surface:
        row = self.data[0]
        rows = row[0]
        columns = row[1]
        surface = GeometricSurface(METRIC['MANHATTAN'])
        logger.info(f'Created surface!')
        return surface



from __future__ import annotations

import logging
from sys import maxsize
from typing import (
    TYPE_CHECKING,
    Set,
    Any,
    Dict,
)
from uuid import (
    uuid4,
)

from .abc import (
    Model,
)

if TYPE_CHECKING:
    from uuid import (
        UUID,
    )
    from .positions import (
        Position,
    )

logger = logging.getLogger(__name__)


class Vehicle(Model):
    initial: Position
    capacity: int
    earliest: float
    timeout: float
    uuid: UUID

    def __init__(self, initial: Position, final: Position = None, capacity: int = 1, earliest: float = 0.0,
                 timeout: float = None, uuid: UUID = None):

        if uuid is None:
            uuid = uuid4()

        self.initial = initial
        self._final = final
        self.capacity = capacity
        self.earliest = earliest
        self.timeout = timeout
        self.uuid = uuid

    @property
    def latest(self) -> float:
        if self.timeout is None:
            return maxsize
        return self.earliest + self.timeout

    @property
    def final(self) -> Position:
        if self._final is None:
            return self.initial
        return self._final

    def as_dict(self) -> Dict[str, Any]:
        return {
            'initial': self.initial,
            'final': self.final,
            'capacity': self.capacity,
            'earliest': self.earliest,
            'timeout': self.timeout,
            'latest': self.latest,
            'uuid': self.uuid,
        }


class Fleet(Model):
    vehicles: Set[Vehicle]

    def __init__(self, vehicles: Set[Vehicle]):
        self.vehicles = vehicles

    def __iter__(self):
        yield from self.vehicles

    def as_dict(self) -> Dict[str, Any]:
        vehicles_str = ', '.join(str(vehicle) for vehicle in self.vehicles)
        dict_values = {
            'vehicles': f'{{{vehicles_str}}}'
        }
        return dict_values



from __future__ import annotations

import logging
from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    TYPE_CHECKING,
)
from uuid import (
    uuid4,
)

if TYPE_CHECKING:
    from typing import (
        List,
        Sequence,
    )
    from uuid import (
        UUID,
    )
    from .surfaces import (
        Surface,
    )

logger = logging.getLogger(__name__)


class Position(ABC):
    uuid: UUID

    def __init__(self, surface: Surface = None, uuid: UUID = None):
        if uuid is None:
            uuid = uuid4()
        self.surface = surface
        self.uuid = uuid

    def distance_to(self, other: Position) -> float:
        return self.surface.distance(self, other)

    def time_to(self, other: Position, now: float) -> float:
        return self.surface.time(self, other, now)

    @abstractmethod
    def is_equal(self, other: Position) -> bool:
        pass


class GeometricPosition(Position):
    coordinates: List[float]

    def __init__(self, coordinates: Sequence[float], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.coordinates = list(coordinates)

    def __str__(self):
        c = ",".join(f"{x:07.3f}" for x in self)
        return f'({c})'

    def __getitem__(self, item):
        return self.coordinates[item]

    def is_equal(self, other: GeometricPosition) -> bool:
        return self.coordinates == other.coordinates



from __future__ import annotations

import logging
from pathlib import (
    Path,
)

from ..models import (
    Job,
    Surface,
    Fleet,
)
from .abc import (
    Loader,
)

logger = logging.getLogger(__name__)


class FileLoader(Loader):

    def __init__(self, file_path: Path, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not isinstance(file_path, Path):
            file_path = Path(file_path)

        self.file_path = file_path

        self._fleet = None
        self._job = None
        self._surface = None

    def _build_formatter(self, force: bool = True):
        if self.formatter is not None and force is False:
            return
        with self.file_path.open() as file:
            data = tuple(tuple(int(v) for v in line.split()) for line in file.readlines())
            self.formatter = self.formatter_cls(data)

    @property
    def fleet(self) -> Fleet:
        if self._fleet is None:
            self._build_formatter()
            self._fleet = self.formatter.fleet(surface=self.surface)
        return self._fleet

    @property
    def job(self) -> Job:
        if self._job is None:
            self._build_formatter()
            self._job = self.formatter.job(surface=self.surface)
        return self._job

    @property
    def surface(self) -> Surface:
        if self._surface is None:
            self._build_formatter()
            self._surface = self.formatter.surface()
        return self._surface



from random import uniform, randint
from typing import Set

import jinete as jit


def generate_one_surface(*args, **kwargs) -> jit.Surface:
    return jit.GeometricSurface(*args, **kwargs)


def generate_surfaces(n: int, *args, **kwargs) -> Set[jit.Surface]:
    return {
        generate_one_surface(*args, **kwargs) for _ in range(n)
    }


def generate_one_position(x_min: float = -100, x_max: float = 100, y_min: float = -100,
                          y_max: float = 100, surface: jit.Surface = None, *args, **kwargs) -> jit.GeometricPosition:
    if surface is None:
        surface = generate_one_surface(*args, **kwargs)
    return jit.GeometricPosition([uniform(x_min, x_max), uniform(y_min, y_max)], surface=surface)


def generate_positions(n: int, surface: jit.Surface = None, *args, **kwargs) -> Set[jit.GeometricPosition]:
    if surface is None:
        surface = generate_one_surface(*args, **kwargs)
    return {
        generate_one_position(surface=surface, *args, **kwargs) for _ in range(n)
    }


def generate_one_trip(identifier: str = None, earliest_min: float = 0, earliest_max: float = 86400,
                      timeout_min: float = 1800,
                      timeout_max: float = 7200, load_time_min: float = 300, load_time_max: float = 900,
                      capacity_min: int = 1, capacity_max: int = 3, *args, **kwargs) -> jit.Trip:
    if identifier is None:
        identifier = str()
    origin, destination = tuple(generate_positions(2, *args, **kwargs))
    earliest = uniform(earliest_min, earliest_max)
    timeout = uniform(timeout_min, timeout_max)
    capacity = randint(capacity_min, capacity_max)
    load_time = uniform(load_time_min, load_time_max)
    return jit.Trip(identifier, origin, destination, earliest, timeout, load_time, capacity)


def generate_trips(n: int, *args, **kwargs) -> Set[jit.Trip]:
    return {
        generate_one_trip(str(i), *args, **kwargs) for i in range(n)
    }


def generate_one_planned_trip(feasible: bool, route=None, *args, **kwargs) -> jit.PlannedTrip:
    trip = generate_one_trip(*args, **kwargs)

    # TODO: Improve feasible randomness.
    if feasible:
        collection_time = trip.earliest
        delivery_time = trip.latest
    else:
        collection_time = trip.earliest - 3600
        delivery_time = trip.latest + 3600

    return jit.PlannedTrip(route, trip, collection_time, delivery_time)


def generate_planned_trips(n: int, *args, **kwargs) -> Set[jit.PlannedTrip]:
    return {
        generate_one_planned_trip(*args, **kwargs) for _ in range(n)
    }


def generate_one_vehicle(capacity_min: int = 1, capacity_max: int = 3, earliest_min: float = 0,
                         earliest_max: float = 86400, timeout_min: float = 14400, timeout_max: float = 28800,
                         *args, **kwargs) -> jit.Vehicle:
    # TODO: Increase parameter options.
    capacity = randint(capacity_min, capacity_max)
    initial = generate_one_position(*args, **kwargs)
    earliest = uniform(earliest_min, earliest_max)
    timeout = uniform(timeout_min, timeout_max)
    return jit.Vehicle(initial, capacity=capacity, earliest=earliest, timeout=timeout)


def generate_vehicles(n: int, *args, **kwargs) -> Set[jit.Vehicle]:
    return {
        generate_one_vehicle(*args, **kwargs) for _ in range(n)
    }


def generate_one_route(feasible: bool, planned_trips_min: int = 1, planned_trips_max: int = 20,
                       *args, **kwargs) -> jit.Route:
    vehicle = generate_one_vehicle(*args, **kwargs)
    surface = generate_one_surface(*args, **kwargs)
    route = jit.Route(vehicle)

    planned_trips_len = randint(planned_trips_min, planned_trips_max)
    cut_len = vehicle.timeout / planned_trips_len

    planned_trip = generate_one_planned_trip(
        feasible=feasible,
        route=route,
        earliest_min=vehicle.earliest,
        earliest_max=vehicle.earliest,
        timeout_min=cut_len,
        timeout_max=cut_len,
        surface=surface,
    )

    route.append_planned_trip(planned_trip)

    for i in range(planned_trips_len):
        cut_min = vehicle.earliest + cut_len * i

        planned_trip = generate_one_planned_trip(
            feasible=feasible,
            route=route,
            earliest_min=route.last_time,
            earliest_max=route.last_time,
            timeout_min=cut_len,
            timeout_max=cut_len,
            surface=surface,
        )

        route.append_planned_trip(planned_trip)
    route.finish()
    vehicle.timeout = route.last_time - vehicle.earliest
    return route


def generate_routes(n: int, *args, **kwargs) -> Set[jit.Route]:
    return {
        generate_one_route(*args, **kwargs) for _ in range(n)
    }



from abc import ABC, abstractmethod
from typing import Type

from ..models import Result

from .formatters import (
    StorerFormatter,
    ColumnarStorerFormatter,
)


class Storer(ABC):

    def __init__(self, result: Result, formatter_cls: Type[StorerFormatter] = None):
        if formatter_cls is None:
            formatter_cls = ColumnarStorerFormatter
        self.result = result
        self.formatter_cls = formatter_cls

    def formatted_planning(self) -> str:
        return self.formatter_cls(result=self.result).format()

    @abstractmethod
    def store(self) -> None:
        pass



from ..models import (
    Planning,
)
from .abc import (
    Algorithm,
)


class NaiveAlgorithm(Algorithm):

    def optimize(self) -> Planning:
        routes = set()
        return Planning(routes)



from .formatters import (
    StorerFormatter,
    ColumnarStorerFormatter,
    HashCodeStorerFormatter,
)

from .abc import (
    Storer,
)
from .file import (
    FileStorer,
)
from .prompt import (
    PromptStorer,
)
from .naive import (
    NaiveStorer,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
from distutils.util import convert_path

module_name = 'jinete'

with open('README.md') as f:
    long_description = f.read()

with open(convert_path('{}/_version.py'.format(module_name))) as file:
    main_ns = dict()
    exec(file.read(), main_ns)
    module_version = main_ns['__version__']

description = (
    "High Performance solving suite for the Pickup and Delivery "
    "Problem and its related extensions."
)
setup(
    name=module_name,
    version=module_version,
    url='https://github.com/garciparedes/jinete',
    author='Sergio García Prado',
    author_email='sergio@garciparedes.me',
    long_description=long_description,
    long_description_content_type='text/markdown',
    description=description,
    packages=find_packages(),
    install_requires=[

    ],
)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import (
        Set,
    )
    from ..models import (
        Trip,
    )


class Result(object):

    def __init__(self, fleet, job, algorithm_cls, planning, computation_time):
        self.fleet = fleet
        self.job = job
        self.algorithm_cls = algorithm_cls
        self.planning = planning
        self.computation_time = computation_time

    @property
    def completed_trips(self) -> Set[Trip]:
        trips = set()
        for route in self.planning.routes:
            trips |= set(route.loaded_trips)
        return trips

    @property
    def coverage_rate(self):
        return len(self.completed_trips) / len(self.job.trips)



from __future__ import annotations
from abc import ABC, abstractmethod

from typing import (
    TYPE_CHECKING,
)

if TYPE_CHECKING:
    from typing import (
        Set,
    )
    from ...models import (
        Result,
        Planning,
        Job,
        Route,
    )


class StorerFormatter(ABC):
    def __init__(self, result: Result, remove_empty_routs: bool = False):
        self.result = result
        self.remove_empty_routes = remove_empty_routs

    @property
    def job(self) -> Job:
        return self.result.job

    @property
    def planning(self) -> Planning:
        return self.result.planning

    @property
    def routes(self) -> Set[Route]:
        if self.remove_empty_routes:
            return self.planning.loaded_routes
        return self.planning.routes

    @abstractmethod
    def format(self) -> str:
        pass
